"""Session derivation from Bloomberg exchange metadata."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from xbbg._core import ext_derive_sessions

if TYPE_CHECKING:
    from xbbg.markets.bloomberg import ExchangeInfo


@dataclass
class SessionWindows:
    """Trading session windows for a security."""

    day: tuple[str, str] | None = None
    allday: tuple[str, str] | None = None
    pre: tuple[str, str] | None = None
    post: tuple[str, str] | None = None
    am: tuple[str, str] | None = None
    pm: tuple[str, str] | None = None

    def to_dict(self) -> dict[str, tuple[str, str]]:
        """Convert to dict, excluding None values."""
        result: dict[str, tuple[str, str]] = {}
        if self.day:
            result["day"] = self.day
        if self.allday:
            result["allday"] = self.allday
        if self.pre:
            result["pre"] = self.pre
        if self.post:
            result["post"] = self.post
        if self.am:
            result["am"] = self.am
        if self.pm:
            result["pm"] = self.pm
        return result


def derive_sessions(exchange_info: ExchangeInfo) -> SessionWindows:
    """Derive session windows from ExchangeInfo using Rust market rules."""
    regular = exchange_info.sessions.get("regular")
    futures = exchange_info.sessions.get("futures")
    base_session = regular or futures
    if not base_session:
        return SessionWindows()

    values = ext_derive_sessions(
        day_start=base_session[0],
        day_end=base_session[1],
        mic=exchange_info.mic,
        exch_code=exchange_info.exch_code,
    )
    return SessionWindows(
        day=values.get("day"),
        allday=values.get("allday"),
        pre=values.get("pre"),
        post=values.get("post"),
        am=values.get("am"),
        pm=values.get("pm"),
    )


def get_session_windows(
    ticker: str,
    mic: str | None = None,
    exch_code: str | None = None,
    regular_hours: tuple[str, str] | None = None,
) -> SessionWindows:
    """Convenience helper to derive sessions without a Bloomberg query."""
    from xbbg.markets.bloomberg import ExchangeInfo

    sessions: dict[str, tuple[str, str]] = {}
    if regular_hours:
        sessions["regular"] = regular_hours

    info = ExchangeInfo(ticker=ticker, mic=mic, exch_code=exch_code, sessions=sessions)
    return derive_sessions(info)



"""Testing helpers for non-live xbbg request and response flows."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
import importlib
from typing import Any

from . import blp

pa = importlib.import_module("pyarrow")


def _nw_module():
    return importlib.import_module("narwhals.stable.v1")


def _require_blpapi():
    try:
        import blpapi  # type: ignore[import-not-found]
    except ImportError as exc:
        raise ImportError(
            "xbbg.testing requires the Bloomberg Python package for TestUtil-backed helpers. "
            "Install `blpapi` to use create_mock_event()/deserialize_service() helpers."
        ) from exc
    return blpapi


def create_mock_event(event_type: int):
    blpapi = _require_blpapi()
    return blpapi.test.createEvent(event_type)


def deserialize_service(service_xml: str):
    blpapi = _require_blpapi()
    return blpapi.test.deserializeService(service_xml)


def get_admin_message_definition(message_name: str | Any):
    blpapi = _require_blpapi()
    if isinstance(message_name, str):
        message_name = blpapi.Name(message_name)
    return blpapi.test.getAdminMessageDefinition(message_name)


def append_message_dict(event, element_def, payload: Mapping[str, Any], properties: Any | None = None):
    blpapi = _require_blpapi()
    formatter = blpapi.test.appendMessage(event, element_def, properties)
    formatter.formatMessageDict(dict(payload))
    return formatter


def _coerce_blpapi_service(service: str | Any, service_xml: str | None):
    if service_xml is not None:
        return deserialize_service(service_xml)

    if hasattr(service, "getOperation"):
        return service
    if not isinstance(service, str):
        raise TypeError("service must be a Bloomberg Service, service URI string, or XML-backed service")

    raise ValueError(
        "A raw service URI cannot be converted to a TestUtil service definition by itself. "
        "Pass `service_xml=` or a pre-built `blpapi.Service` when you need a real mock event."
    )


def _build_message_properties(
    *,
    correlation_ids: Sequence[Any] | None = None,
    request_id: str | None = None,
    service: Any | None = None,
):
    if correlation_ids is None and request_id is None and service is None:
        return None

    blpapi = _require_blpapi()
    props = blpapi.test.MessageProperties()
    if correlation_ids:
        props.setCorrelationIds(list(correlation_ids))
    if request_id:
        props.setRequestId(request_id)
    if service is not None:
        props.setService(service)
    return props


def _stringify_value(value: Any) -> str | None:
    if value is None:
        return None
    return str(value)


def _reference_rows(data: Mapping[str, Mapping[str, Any]]) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    for ticker, fields in data.items():
        for field_name, value in fields.items():
            rows.append(
                {
                    "ticker": ticker,
                    "field": field_name,
                    "value": _stringify_value(value),
                }
            )
    return rows


def _historical_rows(data: Mapping[str, Any]) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    for ticker, ticker_payload in data.items():
        if isinstance(ticker_payload, Mapping):
            for date_value, fields in ticker_payload.items():
                if not isinstance(fields, Mapping):
                    raise TypeError("HistoricalDataRequest mapping values must be field mappings")
                for field_name, value in fields.items():
                    rows.append(
                        {
                            "ticker": ticker,
                            "date": str(date_value),
                            "field": field_name,
                            "value": _stringify_value(value),
                        }
                    )
        elif isinstance(ticker_payload, Sequence) and not isinstance(ticker_payload, (str, bytes, bytearray)):
            for entry in ticker_payload:
                if not isinstance(entry, Mapping):
                    raise TypeError("HistoricalDataRequest row entries must be mappings")
                date_value = entry.get("date")
                for key, value in entry.items():
                    if key == "date":
                        continue
                    rows.append(
                        {
                            "ticker": ticker,
                            "date": str(date_value),
                            "field": key,
                            "value": _stringify_value(value),
                        }
                    )
        else:
            raise TypeError("Unsupported HistoricalDataRequest payload shape")
    return rows


def _rows_from_operation(operation: str, data: Any) -> list[dict[str, Any]]:
    if isinstance(data, Sequence) and not isinstance(data, (str, bytes, bytearray)):
        rows = [dict(row) for row in data]
        if not all(isinstance(row, Mapping) for row in data):
            raise TypeError("Explicit mock rows must be mappings")
        return rows

    if not isinstance(data, Mapping):
        raise TypeError("Mock response data must be a mapping or a list of row mappings")

    if operation == "ReferenceDataRequest":
        return _reference_rows(data)
    if operation == "HistoricalDataRequest":
        return _historical_rows(data)

    raise ValueError(
        f"No default row builder for operation {operation!r}. Pass an explicit list of row mappings for this operation."
    )


def _table_from_rows(rows: Sequence[Mapping[str, Any]]) -> Any:
    if not rows:
        return pa.table({})
    return pa.Table.from_pylist([dict(row) for row in rows])


@dataclass(slots=True)
class MockResponse:
    """Container for a canned xbbg response and optional Bloomberg event."""

    service: str
    operation: str
    table: Any
    event: Any | None = None
    request_operation: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)

    def matches(self, params_dict: Mapping[str, Any]) -> bool:
        if params_dict.get("service") != self.service:
            return False

        actual_operation = params_dict.get("request_operation") or params_dict.get("operation")
        expected_operation = self.request_operation or self.operation
        return actual_operation == expected_operation


def create_mock_response(
    *,
    service: str | Any,
    operation: str,
    data: Any,
    service_xml: str | None = None,
    event_type: int | None = None,
    correlation_ids: Sequence[Any] | None = None,
    request_id: str | None = None,
) -> MockResponse:
    if isinstance(service, str):
        service_name = service
    else:
        service_name = service.name()
    rows = _rows_from_operation(operation, data)
    table = _table_from_rows(rows)

    event = None
    service_obj: Any | None = None
    if service_xml is not None:
        service_obj = _coerce_blpapi_service(service, service_xml)
    elif not isinstance(service, str):
        service_obj = service

    if service_obj is not None:
        op = service_obj.getOperation(operation)
        element_def = op.getResponseDefinitionAt(0)
        event = create_mock_event(event_type or _require_blpapi().Event.RESPONSE)
        props = _build_message_properties(
            correlation_ids=correlation_ids,
            request_id=request_id,
            service=service_obj,
        )
        append_message_dict(event, element_def, data, props)

    return MockResponse(
        service=service_name,
        operation=operation,
        table=table,
        event=event,
        metadata={"rows": rows},
    )


def _coerce_mock_table(value: MockResponse | Any | Sequence[Mapping[str, Any]]) -> Any:
    if isinstance(value, MockResponse):
        return value.table
    if isinstance(value, pa.Table):
        return value
    if isinstance(value, pa.RecordBatch):
        return pa.Table.from_batches([value])
    if isinstance(value, Sequence) and not isinstance(value, (str, bytes, bytearray)):
        return _table_from_rows(value)
    raise TypeError("Unsupported mock response value")


class mock_engine:
    """Middleware context manager that serves canned responses."""

    def __init__(
        self,
        responses: Sequence[MockResponse | Any | Sequence[Mapping[str, Any]]],
        *,
        strict: bool = True,
    ) -> None:
        self._responses = list(responses)
        self._strict = strict
        self._middleware = None

    def _pop_match(self, params_dict: Mapping[str, Any]):
        for index, candidate in enumerate(self._responses):
            if isinstance(candidate, MockResponse):
                if candidate.matches(params_dict):
                    return self._responses.pop(index)
            else:
                if index == 0:
                    return self._responses.pop(index)
        return None

    async def _handle(self, context: blp.RequestContext, call_next):
        match = self._pop_match(context.params_dict)
        if match is None:
            if self._strict:
                raise LookupError(f"No mock response matched {context.params.service}::{context.params.operation}")
            return await call_next(context)

        table = _coerce_mock_table(match)
        batch = next(iter(table.to_batches()), None)
        if batch is None:
            batch = pa.record_batch([], names=[])

        context.metadata["mocked"] = True
        if isinstance(match, MockResponse):
            context.metadata["mock_response"] = match
        context.batch = batch
        context.table = table
        context.elapsed_ms = 0.0

        nw_df = _nw_module().from_native(table)
        context.frame = blp._convert_backend(nw_df, context.backend)
        return context.frame

    def __enter__(self):
        async def middleware(context: blp.RequestContext, call_next):
            return await self._handle(context, call_next)

        self._middleware = middleware
        blp.add_middleware(middleware)
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        if self._middleware is not None:
            try:
                blp.remove_middleware(self._middleware)
            finally:
                self._middleware = None


__all__ = [
    "MockResponse",
    "append_message_dict",
    "create_mock_event",
    "create_mock_response",
    "deserialize_service",
    "get_admin_message_definition",
    "mock_engine",
]



"""Extension functions for xbbg.

This module contains convenience wrappers built on top of the core API.
These are pure Python functions that compose core operations (bdp, bds, bdh, bql)
for common use cases.

Extension Categories:
    - historical: dividend(), earnings(), turnover(), etf_holdings()
    - futures: fut_ticker(), active_futures(), cdx_ticker(), active_cdx()
    - currency: convert_ccy()
    - fixed_income: yas(), preferreds(), corporate_bonds(), bqr()
    - bonds: bond_info(), bond_risk(), bond_spreads(), bond_cashflows(), bond_key_rates(), bond_curve()
    - options: option_info(), option_greeks(), option_pricing(), option_chain(), option_chain_bql(), option_screen()
    - cdx: cdx_info(), cdx_defaults(), cdx_pricing(), cdx_risk(), cdx_basis(), cdx_default_prob(), cdx_cashflows(), cdx_curve()

Async versions (primary implementations):
    - historical: adividend(), aearnings(), aturnover(), aetf_holdings()
    - futures: afut_ticker(), aactive_futures(), acdx_ticker(), aactive_cdx()
    - currency: aconvert_ccy()
    - fixed_income: ayas(), apreferreds(), acorporate_bonds(), abqr()
    - bonds: abond_info(), abond_risk(), abond_spreads(), abond_cashflows(), abond_key_rates(), abond_curve()
    - options: aoption_info(), aoption_greeks(), aoption_pricing(), aoption_chain(), aoption_chain_bql(), aoption_screen()
    - cdx: acdx_info(), acdx_defaults(), acdx_pricing(), acdx_risk(), acdx_basis(), acdx_default_prob(), acdx_cashflows(), acdx_curve()

Example::

    from xbbg import ext

    # Get dividend history
    df = ext.dividend("AAPL US Equity")

    # Get ETF holdings
    df = ext.etf_holdings("SPY US Equity")

    # Resolve futures ticker
    ticker = ext.fut_ticker("ES1 Index", "2024-01-15")

    # Convert currency
    df_usd = ext.convert_ccy(df, ccy="USD")

    # Yield & spread analysis for bonds
    df = ext.yas("US912810TM69 Govt", "YAS_BOND_YLD")

    # Find preferred stocks
    df = ext.preferreds("BAC US Equity")

    # Find corporate bonds
    df = ext.corporate_bonds("AAPL")

    # Bond analytics
    df = ext.bond_info("T 4.5 05/15/38 Govt")
    df = ext.bond_risk("T 4.5 05/15/38 Govt")
    df = ext.bond_spreads("T 4.5 05/15/38 Govt")

    # Options analytics
    df = ext.option_info("AAPL US 01/17/25 C200 Equity")
    df = ext.option_greeks("AAPL US 01/17/25 C200 Equity")
    chain = ext.option_chain("AAPL US Equity")

    # CDX analytics
    df = ext.cdx_info("CDX IG CDSI GEN 5Y Corp")
    df = ext.cdx_pricing("CDX IG CDSI GEN 5Y Corp")

    # Async example
    import asyncio


    async def main():
        df = await ext.adividend("AAPL US Equity")
        print(df)


    asyncio.run(main())
"""

from __future__ import annotations

# Sync functions
# Async functions
from xbbg.ext.bonds import (
    abond_cashflows,
    abond_curve,
    abond_info,
    abond_key_rates,
    abond_risk,
    abond_spreads,
    bond_cashflows,
    bond_curve,
    bond_info,
    bond_key_rates,
    bond_risk,
    bond_spreads,
)
from xbbg.ext.cdx import (
    acdx_basis,
    acdx_cashflows,
    acdx_curve,
    acdx_default_prob,
    acdx_defaults,
    acdx_info,
    acdx_pricing,
    acdx_risk,
    cdx_basis,
    cdx_cashflows,
    cdx_curve,
    cdx_default_prob,
    cdx_defaults,
    cdx_info,
    cdx_pricing,
    cdx_risk,
)
from xbbg.ext.currency import aconvert_ccy, convert_ccy
from xbbg.ext.fixed_income import (
    YieldType,
    abqr,
    acorporate_bonds,
    apreferreds,
    ayas,
    bqr,
    corporate_bonds,
    preferreds,
    yas,
)
from xbbg.ext.futures import (
    aactive_cdx,
    aactive_futures,
    acdx_ticker,
    active_cdx,
    active_futures,
    afut_ticker,
    cdx_ticker,
    fut_ticker,
)
from xbbg.ext.historical import (
    adividend,
    aearnings,
    aetf_holdings,
    aturnover,
    dividend,
    earnings,
    etf_holdings,
    turnover,
)
from xbbg.ext.options import (
    ChainPeriodicity,
    ExerciseType,
    ExpiryMatch,
    PutCall,
    StrikeRef,
    aoption_chain,
    aoption_chain_bql,
    aoption_greeks,
    aoption_info,
    aoption_pricing,
    aoption_screen,
    option_chain,
    option_chain_bql,
    option_greeks,
    option_info,
    option_pricing,
    option_screen,
)

__all__ = [
    # Historical extensions (sync)
    "dividend",
    "earnings",
    "turnover",
    "etf_holdings",
    # Historical extensions (async)
    "adividend",
    "aearnings",
    "aturnover",
    "aetf_holdings",
    # Futures extensions (sync)
    "fut_ticker",
    "active_futures",
    "cdx_ticker",
    "active_cdx",
    # Futures extensions (async)
    "afut_ticker",
    "aactive_futures",
    "acdx_ticker",
    "aactive_cdx",
    # Currency extensions (sync)
    "convert_ccy",
    # Currency extensions (async)
    "aconvert_ccy",
    # Fixed income extensions (sync)
    "yas",
    "YieldType",
    "preferreds",
    "corporate_bonds",
    "bqr",
    # Fixed income extensions (async)
    "ayas",
    "apreferreds",
    "acorporate_bonds",
    "abqr",
    # Bond analytics (sync)
    "bond_info",
    "bond_risk",
    "bond_spreads",
    "bond_cashflows",
    "bond_key_rates",
    "bond_curve",
    # Bond analytics (async)
    "abond_info",
    "abond_risk",
    "abond_spreads",
    "abond_cashflows",
    "abond_key_rates",
    "abond_curve",
    # Options analytics enums
    "PutCall",
    "ChainPeriodicity",
    "StrikeRef",
    "ExerciseType",
    "ExpiryMatch",
    # Options analytics (sync)
    "option_info",
    "option_greeks",
    "option_pricing",
    "option_chain",
    "option_chain_bql",
    "option_screen",
    # Options analytics (async)
    "aoption_info",
    "aoption_greeks",
    "aoption_pricing",
    "aoption_chain",
    "aoption_chain_bql",
    "aoption_screen",
    # CDX analytics (sync)
    "cdx_info",
    "cdx_defaults",
    "cdx_pricing",
    "cdx_risk",
    "cdx_basis",
    "cdx_default_prob",
    "cdx_cashflows",
    "cdx_curve",
    # CDX analytics (async)
    "acdx_info",
    "acdx_defaults",
    "acdx_pricing",
    "acdx_risk",
    "acdx_basis",
    "acdx_default_prob",
    "acdx_cashflows",
    "acdx_curve",
]



"""Bloomberg exchange metadata query module."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from datetime import datetime
import importlib
import logging
import re
from typing import Any, cast

import narwhals.stable.v1 as nw
import pandas as pd

logger = logging.getLogger(__name__)

EXCHANGE_FIELDS = [
    "IANA_TIME_ZONE",
    "TIME_ZONE_NUM",
    "ID_MIC_PRIM_EXCH",
    "EXCH_CODE",
    "COUNTRY_ISO",
    "TRADING_DAY_START_TIME_EOD",
    "TRADING_DAY_END_TIME_EOD",
    "FUT_TRADING_HRS",
]

# Kept for backward compatibility; timezone inference now comes from Rust.
COUNTRY_TIMEZONE_MAP: dict[str, str] = {}


@dataclass
class ExchangeInfo:
    """Exchange metadata from Bloomberg."""

    ticker: str
    mic: str | None = None
    exch_code: str | None = None
    timezone: str = "UTC"
    utc_offset: float | None = None
    sessions: dict[str, tuple[str, str]] = field(default_factory=dict)
    source: str = "fallback"
    cached_at: datetime | None = None


def _parse_hhmm(time_str: str | None) -> str | None:
    """Parse Bloomberg time format (HHMM or HH:MM) to HH:MM."""
    if not time_str or pd.isna(time_str):
        return None

    time_str = str(time_str).strip()
    if not time_str:
        return None

    if ":" in time_str:
        parts = time_str.split(":")
        if len(parts) >= 2:
            try:
                hh = int(parts[0])
                mm = int(parts[1])
                return f"{hh:02d}:{mm:02d}"
            except ValueError:
                return None

    if time_str.isdigit() and len(time_str) == 4:
        try:
            hh = int(time_str[:2])
            mm = int(time_str[2:])
            return f"{hh:02d}:{mm:02d}"
        except ValueError:
            return None

    return None


def _parse_futures_hours(fut_hours: str | None) -> dict[str, tuple[str, str]]:
    """Parse Bloomberg FUT_TRADING_HRS to a session dict."""
    if not fut_hours or pd.isna(fut_hours):
        return {}

    fut_hours = str(fut_hours).strip()
    if not fut_hours:
        return {}

    pattern = r"(\d{1,2}:\d{2}|\d{4})\s*-\s*(\d{1,2}:\d{2}|\d{4})"
    match = re.match(pattern, fut_hours)
    if not match:
        return {}

    start = _parse_hhmm(match.group(1))
    end = _parse_hhmm(match.group(2))
    if start and end:
        return {"futures": (start, end)}
    return {}


def _convert_est_to_local(time_str: str, local_tz: str) -> str:
    """Convert a time string from EST to local timezone."""
    if local_tz in ("America/New_York", "US/Eastern", "EST", "EDT"):
        return time_str

    try:
        ref_date = datetime.now().strftime("%Y-%m-%d")
        est_ts = pd.Timestamp(f"{ref_date} {time_str}", tz="America/New_York")
        local_ts = est_ts.tz_convert(local_tz)
        return f"{local_ts.hour:02d}:{local_ts.minute:02d}"
    except Exception as e:
        logger.debug("Failed to convert time %s from EST to %s: %s", time_str, local_tz, e)
        return time_str


def _parse_trading_hours(
    start_time: str | None,
    end_time: str | None,
    fut_hours: str | None,
    local_tz: str = "America/New_York",
) -> dict[str, tuple[str, str]]:
    """Parse Bloomberg trading hours fields into sessions dict."""
    sessions: dict[str, tuple[str, str]] = {}

    start = _parse_hhmm(start_time)
    end = _parse_hhmm(end_time)
    if start and end:
        local_start = _convert_est_to_local(start, local_tz)
        local_end = _convert_est_to_local(end, local_tz)
        sessions["regular"] = (local_start, local_end)

    sessions.update(_parse_futures_hours(fut_hours))
    return sessions


def _infer_timezone_from_country(country_iso: str | None) -> str | None:
    """Infer IANA timezone from country ISO using Rust fallback map."""
    if not country_iso or pd.isna(country_iso):
        return None
    core = importlib.import_module("xbbg._core")
    infer = cast("Any", getattr(core, "ext_infer_timezone", None))
    if infer is None:
        return None
    return cast("str | None", infer(str(country_iso).strip().upper()))


def _extract_value(df: pd.DataFrame, field: str) -> str | float | None:
    """Extract a single value from Bloomberg response DataFrame."""
    if df.empty:
        return None

    col_map = {c.lower(): c for c in df.columns}
    actual_col = col_map.get(field.lower())
    if actual_col is None:
        return None

    val = df.iloc[0][actual_col]
    if pd.isna(val):
        return None
    return val


def _build_exchange_info_from_response(ticker: str, df: pd.DataFrame) -> ExchangeInfo:
    """Build ExchangeInfo from Bloomberg response DataFrame."""
    if df.empty:
        logger.warning("Empty response from Bloomberg for ticker %s, using fallback", ticker)
        return ExchangeInfo(ticker=ticker, source="fallback")

    iana_tz = _extract_value(df, "IANA_TIME_ZONE")
    tz_num = _extract_value(df, "TIME_ZONE_NUM")
    mic = _extract_value(df, "ID_MIC_PRIM_EXCH")
    exch_code = _extract_value(df, "EXCH_CODE")
    country_iso = _extract_value(df, "COUNTRY_ISO")
    start_time = _extract_value(df, "TRADING_DAY_START_TIME_EOD")
    end_time = _extract_value(df, "TRADING_DAY_END_TIME_EOD")
    fut_hours = _extract_value(df, "FUT_TRADING_HRS")

    timezone = "UTC"
    source = "bloomberg"
    if iana_tz and isinstance(iana_tz, str):
        timezone = iana_tz
        source = "bloomberg"
    elif country_iso:
        inferred_tz = _infer_timezone_from_country(str(country_iso))
        if inferred_tz:
            timezone = inferred_tz
            source = "inferred"
        else:
            source = "fallback"
    else:
        source = "fallback"

    sessions = _parse_trading_hours(
        str(start_time) if start_time else None,
        str(end_time) if end_time else None,
        str(fut_hours) if fut_hours else None,
        local_tz=timezone,
    )

    return ExchangeInfo(
        ticker=ticker,
        mic=str(mic) if mic else None,
        exch_code=str(exch_code) if exch_code else None,
        timezone=timezone,
        utc_offset=float(tz_num) if tz_num is not None else None,
        sessions=sessions,
        source=source,
        cached_at=None,
    )


def _to_pandas_wide(data: Any) -> pd.DataFrame:
    """Convert current backend response into a wide pandas DataFrame."""
    utils = importlib.import_module("xbbg.ext._utils")
    pivot_fn = cast("Any", utils._pivot_bdp_to_wide)
    nw_df = nw.from_native(data)
    nw_df = pivot_fn(nw_df)
    return nw_df.to_pandas()


async def afetch_exchange_info(ticker: str, **kwargs) -> ExchangeInfo:
    """Async fetch exchange metadata from Bloomberg."""
    xbbg_module = importlib.import_module("xbbg")
    abdp_fn = getattr(xbbg_module, "abdp")  # noqa: B009

    try:
        data = await abdp_fn(tickers=ticker, flds=EXCHANGE_FIELDS, **kwargs)
        return _build_exchange_info_from_response(ticker, _to_pandas_wide(data))
    except Exception as e:
        logger.warning("Failed to fetch exchange info from Bloomberg for %s: %s", ticker, e)
        return ExchangeInfo(ticker=ticker, source="fallback")


def fetch_exchange_info(ticker: str, **kwargs) -> ExchangeInfo:
    """Sync fetch exchange metadata from Bloomberg."""
    return asyncio.run(afetch_exchange_info(ticker=ticker, **kwargs))



"""Market information utilities."""

from __future__ import annotations

from xbbg.markets.bloomberg import ExchangeInfo, afetch_exchange_info, fetch_exchange_info
from xbbg.markets.info import (
    ccy_pair,
    convert_session_times_to_utc,
    exch_info,
    market_info,
    market_timing,
)
from xbbg.markets.overrides import (
    clear_exchange_override,
    get_exchange_override,
    has_override,
    list_exchange_overrides,
    set_exchange_override,
)
from xbbg.markets.sessions import SessionWindows, derive_sessions, get_session_windows

__all__ = [
    "SessionWindows",
    "derive_sessions",
    "get_session_windows",
    "exch_info",
    "market_info",
    "market_timing",
    "ccy_pair",
    "convert_session_times_to_utc",
    "set_exchange_override",
    "get_exchange_override",
    "clear_exchange_override",
    "list_exchange_overrides",
    "has_override",
    "ExchangeInfo",
    "fetch_exchange_info",
    "afetch_exchange_info",
]



from __future__ import annotations

SDK_EXPORTS = (
    "get_sdk_info",
    "set_sdk_path",
    "clear_sdk_path",
)

CORE_EXPORTS = (
    "set_log_level",
    "get_log_level",
    "enable_sdk_logging",
)

SERVICE_EXPORTS = (
    "Service",
    "Operation",
    "OutputMode",
    "RequestParams",
    "ExtractorHint",
)

FIELD_CACHE_EXPORTS = (
    "FieldTypeCache",
    "FieldInfo",
    "resolve_field_types",
    "aresolve_field_types",
    "cache_field_types",
    "get_field_info",
    "get_field_cache_stats",
    "clear_field_cache",
)

SCHEMA_EXPORTS = (
    "get_schema",
    "aget_schema",
    "get_operation",
    "aget_operation",
    "list_operations",
    "alist_operations",
    "get_enum_values",
    "aget_enum_values",
    "list_valid_elements",
    "alist_valid_elements",
    "generate_stubs",
    "configure_ide_stubs",
    "ServiceSchema",
    "OperationSchema",
)

SCHEMA_LOOKUP_EXPORTS = (*SCHEMA_EXPORTS, "ElementInfo")

EXCEPTION_EXPORTS = (
    "BlpError",
    "BlpSessionError",
    "BlpRequestError",
    "BlpSecurityError",
    "BlpFieldError",
    "BlpValidationError",
    "BlpTimeoutError",
    "BlpInternalError",
    "BlpBPipeError",
)

MODULE_EXPORTS = (
    "ext",
    "markets",
    "testing",
)

BACKEND_EXPORTS = (
    "is_backend_available",
    "check_backend",
    "get_available_backends",
    "print_backend_status",
    "is_format_supported",
    "get_supported_formats",
    "check_format_compatibility",
    "validate_backend_format",
)

PACKAGE_GENERIC_EXPORTS = (
    "arequest",
    "request",
)

PACKAGE_SYNC_EXPORTS = (
    "bdp",
    "bds",
    "bdh",
    "bdib",
    "bdtick",
    "bql",
    "bsrch",
    "bflds",
    "beqs",
    "blkp",
    "bport",
    "bcurves",
    "bgovts",
)

PACKAGE_ASYNC_EXPORTS = (
    "abdp",
    "abds",
    "abdh",
    "abdib",
    "abdtick",
    "abql",
    "absrch",
    "abflds",
    "abeqs",
    "ablkp",
    "abport",
    "abcurves",
    "abgovts",
)

PACKAGE_STREAMING_EXPORTS = (
    "Tick",
    "Subscription",
    "asubscribe",
    "subscribe",
    "astream",
    "stream",
    "avwap",
    "vwap",
    "amktbar",
    "mktbar",
    "adepth",
    "depth",
    "achains",
    "chains",
)

PACKAGE_TA_EXPORTS = (
    "abta",
    "bta",
    "ta_studies",
    "ta_study_params",
    "generate_ta_stubs",
)

PACKAGE_CONFIG_EXPORTS = (
    "configure",
    "set_backend",
    "get_backend",
)

PACKAGE_MIDDLEWARE_EXPORTS = (
    "RequestEnvironment",
    "RequestContext",
    "add_middleware",
    "remove_middleware",
    "clear_middleware",
    "get_middleware",
    "set_middleware",
)

PACKAGE_LIFECYCLE_EXPORTS = (
    "shutdown",
    "reset",
    "is_connected",
)

PACKAGE_BLP_SCHEMA_EXPORTS = (
    "bops",
    "abops",
    "bschema",
    "abschema",
)

PACKAGE_BLP_EXPORTS = (
    "Backend",
    *PACKAGE_GENERIC_EXPORTS,
    *PACKAGE_SYNC_EXPORTS,
    *PACKAGE_ASYNC_EXPORTS,
    *PACKAGE_STREAMING_EXPORTS,
    *PACKAGE_TA_EXPORTS,
    *PACKAGE_CONFIG_EXPORTS,
    *PACKAGE_MIDDLEWARE_EXPORTS,
    *PACKAGE_LIFECYCLE_EXPORTS,
    *PACKAGE_BLP_SCHEMA_EXPORTS,
)

PACKAGE_EXPORTS = (
    "__version__",
    "_core",
    "Backend",
    "Engine",
    "EngineConfig",
    *PACKAGE_GENERIC_EXPORTS,
    *PACKAGE_SYNC_EXPORTS,
    *PACKAGE_ASYNC_EXPORTS,
    *PACKAGE_STREAMING_EXPORTS,
    *PACKAGE_TA_EXPORTS,
    *PACKAGE_CONFIG_EXPORTS,
    *PACKAGE_MIDDLEWARE_EXPORTS,
    *PACKAGE_LIFECYCLE_EXPORTS,
    *CORE_EXPORTS,
    *PACKAGE_BLP_SCHEMA_EXPORTS,
    *SDK_EXPORTS,
    *FIELD_CACHE_EXPORTS,
    *SERVICE_EXPORTS,
    *SCHEMA_EXPORTS,
    *EXCEPTION_EXPORTS,
    *MODULE_EXPORTS,
    *BACKEND_EXPORTS,
)

BLP_MODULE_EXPORTS = (
    "Backend",
    "arequest",
    "request",
    "abdp",
    "abdh",
    "abds",
    "abdib",
    "abdtick",
    "abql",
    "absrch",
    "abeqs",
    "ablkp",
    "abport",
    "abcurves",
    "abgovts",
    "abqr",
    "bqr",
    "abflds",
    "bflds",
    "abfld",
    "bfld",
    "afieldInfo",
    "fieldInfo",
    "afieldSearch",
    "fieldSearch",
    "bdp",
    "bdh",
    "bds",
    "bdib",
    "bdtick",
    "bql",
    "bsrch",
    "beqs",
    "blkp",
    "bport",
    "bcurves",
    "bgovts",
    "Tick",
    "Subscription",
    "asubscribe",
    "subscribe",
    "astream",
    "stream",
    "avwap",
    "vwap",
    "amktbar",
    "mktbar",
    "adepth",
    "depth",
    "achains",
    "chains",
    "abta",
    "bta",
    "ta_studies",
    "ta_study_params",
    "generate_ta_stubs",
    "configure",
    "set_backend",
    "get_backend",
    "RequestContext",
    "add_middleware",
    "remove_middleware",
    "clear_middleware",
    "get_middleware",
    "set_middleware",
    "Service",
    "Operation",
    "OutputMode",
    "RequestParams",
    "ExtractorHint",
    "abops",
    "bops",
    "abschema",
    "bschema",
)



"""Bloomberg service definitions and request parameters.

Enum definitions (Service, Operation, ExtractorHint, Format, OutputMode) are
generated from ``defs/bloomberg.toml`` into ``_services_gen.py``.  This module
re-exports them and provides the hand-written :class:`RequestParams` dataclass.

Example::

    from xbbg import Service, Operation, RequestParams

    params = RequestParams(
        service=Service.REFDATA,
        operation=Operation.REFERENCE_DATA,
        securities=["AAPL US Equity"],
        fields=["PX_LAST"],
    )
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

from xbbg._services_gen import (
    ExtractorHint,
    Format,
    Operation,
    OutputMode,
    Service,
)
from xbbg.exceptions import BlpValidationError

# Backwards compatibility alias
LongMode = Format


@dataclass
class RequestParams:
    """Validated request parameters for the Bloomberg API.

    This dataclass holds all possible parameters for Bloomberg requests.
    Not all parameters are used for all request types - the Python layer
    validates that required parameters are present for each operation.

    The Rust layer handles default extractor resolution via
    ``RequestParams::with_defaults()``, so Python only passes an explicit
    extractor when the caller overrides it.

    Attributes:
        service: Bloomberg service URI (e.g., ``"//blp/refdata"``).
        operation: Request operation name (e.g., ``"ReferenceDataRequest"``).
        request_operation: Actual Bloomberg operation name when using
            ``Operation.RAW_REQUEST`` as the low-level escape hatch.
        securities: List of security identifiers (for multi-security requests).
        security: Single security identifier (for intraday requests).
        fields: List of field names to retrieve.
        overrides: List of (field, value) tuples for field overrides.
        elements: List of (name, value) tuples for generic request elements (BQL, bsrch).
        start_date: Start date for historical requests (YYYYMMDD format).
        end_date: End date for historical requests (YYYYMMDD format).
        start_datetime: Start datetime for intraday requests (ISO format).
        end_datetime: End datetime for intraday requests (ISO format).
        request_tz: How naive intraday datetimes are interpreted before the API call
            (``UTC``, ``local``, ``exchange``, ``NY``/``LN``/…, reference ticker, or IANA).
        output_tz: Relabel Arrow ``time`` to this zone (same instants; handled in Rust).
        event_type: Event type for intraday bars (TRADE, BID, ASK, etc.).
        event_types: Event types for intraday ticks (TRADE, BID, ASK, etc.).
        interval: Bar interval in minutes for intraday bars.
        options: Additional Bloomberg options as (key, value) tuples.
        field_types: Manual type overrides for fields (for issue #168).
        output: Output format (arrow or json).
        extractor: Override the auto-detected extractor hint.  When ``None``
            the Rust layer picks the correct extractor for the operation.
        format: Output format (LONG, LONG_TYPED, LONG_WITH_METADATA).
        include_security_errors: When True for ReferenceData requests, include
            ``__SECURITY_ERROR__`` rows for securities that failed.
        validate_fields: Optional per-request override for field validation.
            ``True`` forces strict validation, ``False`` disables it, and
            ``None`` (default) follows engine configuration.
    """

    service: str | Service
    operation: str | Operation
    request_operation: str | Operation | None = None
    securities: Sequence[str] | None = None
    security: str | None = None
    fields: Sequence[str] | None = None
    overrides: Sequence[tuple[str, str]] | None = None
    elements: Sequence[tuple[str, str]] | None = None
    start_date: str | None = None
    end_date: str | None = None
    start_datetime: str | None = None
    end_datetime: str | None = None
    request_tz: str | None = None
    output_tz: str | None = None
    event_type: str | None = None
    event_types: Sequence[str] | None = None
    interval: int | None = None
    options: Sequence[tuple[str, str]] | None = None
    field_types: dict[str, str] | None = None
    output: OutputMode = OutputMode.ARROW
    extractor: ExtractorHint | None = None
    format: Format | None = None
    include_security_errors: bool = False
    validate_fields: bool | None = None

    def __post_init__(self) -> None:
        """Convert enums to strings and set defaults."""
        if isinstance(self.service, Service):
            self.service = self.service.value
        if isinstance(self.operation, Operation):
            self.operation = self.operation.value
        if isinstance(self.request_operation, Operation):
            self.request_operation = self.request_operation.value
        if isinstance(self.output, str):
            self.output = OutputMode(self.output)

    def validate(self) -> None:
        """Validate parameters for the given operation.

        Raises:
            BlpValidationError: If required parameters are missing or invalid.
        """
        if not self.service:
            raise BlpValidationError("service is required")

        if self.operation == Operation.RAW_REQUEST.value:
            if not self.request_operation:
                raise BlpValidationError("request_operation is required for RawRequest")
            return

        if not self.operation:
            raise BlpValidationError("operation is required")

        op = self.operation

        # Validate based on operation type
        if op == Operation.REFERENCE_DATA.value:
            self._validate_reference_data()
        elif op == Operation.HISTORICAL_DATA.value:
            self._validate_historical_data()
        elif op == Operation.INTRADAY_BAR.value:
            self._validate_intraday_bar()
        elif op == Operation.INTRADAY_TICK.value:
            self._validate_intraday_tick()
        elif op in (Operation.FIELD_INFO.value, Operation.FIELD_SEARCH.value):
            self._validate_field_request()
        # Unknown operations: no validation (power user mode)

    def _validate_reference_data(self) -> None:
        """Validate ReferenceDataRequest parameters."""
        if not self.securities:
            raise BlpValidationError("securities is required for ReferenceDataRequest")
        if not self.fields:
            raise BlpValidationError("fields is required for ReferenceDataRequest")

    def _validate_historical_data(self) -> None:
        """Validate HistoricalDataRequest parameters."""
        if not self.securities:
            raise BlpValidationError("securities is required for HistoricalDataRequest")
        if not self.fields:
            raise BlpValidationError("fields is required for HistoricalDataRequest")
        if not self.start_date:
            raise BlpValidationError("start_date is required for HistoricalDataRequest")
        if not self.end_date:
            raise BlpValidationError("end_date is required for HistoricalDataRequest")

    def _validate_intraday_bar(self) -> None:
        """Validate IntradayBarRequest parameters."""
        if not self.security:
            raise BlpValidationError("security is required for IntradayBarRequest")
        if not self.event_type:
            raise BlpValidationError("event_type is required for IntradayBarRequest")
        if self.interval is None:
            raise BlpValidationError("interval is required for IntradayBarRequest")
        if not self.start_datetime:
            raise BlpValidationError("start_datetime is required for IntradayBarRequest")
        if not self.end_datetime:
            raise BlpValidationError("end_datetime is required for IntradayBarRequest")

    def _validate_intraday_tick(self) -> None:
        """Validate IntradayTickRequest parameters."""
        if not self.security:
            raise BlpValidationError("security is required for IntradayTickRequest")
        if not self.start_datetime:
            raise BlpValidationError("start_datetime is required for IntradayTickRequest")
        if not self.end_datetime:
            raise BlpValidationError("end_datetime is required for IntradayTickRequest")

    def _validate_field_request(self) -> None:
        """Validate FieldInfoRequest/FieldSearchRequest parameters."""
        if not self.fields:
            raise BlpValidationError("fields is required for field metadata requests")

    def to_dict(self) -> dict[str, object]:
        """Convert to dictionary for passing to Rust.

        Only non-``None`` values are included.  When *extractor* is ``None``
        the key is omitted so the Rust layer can apply its own default via
        ``RequestParams::with_defaults()``.

        Returns:
            Dictionary suitable for Rust consumption.
        """
        result: dict[str, object] = {
            "service": self.service,
            "operation": self.operation,
        }

        if self.request_operation is not None:
            result["request_operation"] = self.request_operation

        # Only pass extractor when explicitly set by the caller
        if self.extractor is not None:
            result["extractor"] = self.extractor.value

        if self.securities is not None:
            result["securities"] = list(self.securities)
        if self.security is not None:
            result["security"] = self.security
        if self.fields is not None:
            # FieldInfoRequest uses "id" array in Bloomberg API, mapped to field_ids in Rust
            if self.operation == Operation.FIELD_INFO.value:
                result["field_ids"] = list(self.fields)
            # FieldSearchRequest uses search_spec (single string)
            elif self.operation == Operation.FIELD_SEARCH.value:
                result["search_spec"] = self.fields[0] if self.fields else ""
            else:
                result["fields"] = list(self.fields)
        if self.overrides is not None:
            result["overrides"] = list(self.overrides)
        if self.elements is not None:
            result["elements"] = list(self.elements)
        if self.start_date is not None:
            result["start_date"] = self.start_date
        if self.end_date is not None:
            result["end_date"] = self.end_date
        if self.start_datetime is not None:
            result["start_datetime"] = self.start_datetime
        if self.end_datetime is not None:
            result["end_datetime"] = self.end_datetime
        if self.request_tz is not None:
            result["request_tz"] = self.request_tz
        if self.output_tz is not None:
            result["output_tz"] = self.output_tz
        if self.event_type is not None:
            result["event_type"] = self.event_type
        if self.event_types is not None:
            result["event_types"] = list(self.event_types)
        if self.interval is not None:
            result["interval"] = self.interval
        if self.options is not None:
            result["options"] = list(self.options)
        if self.field_types is not None:
            result["field_types"] = self.field_types
        if self.format is not None:
            # Pass format value to Rust
            result["format"] = self.format.value if isinstance(self.format, Format) else self.format
        if self.include_security_errors:
            result["include_security_errors"] = True
        if self.validate_fields is not None:
            result["validate_fields"] = self.validate_fields

        return result
